#include <string.h>
#include <dos.h>
#include "x_err.hpp"
#include "x_text.hpp"
#include "x_comm.hpp"

#define CLI()		_disable()
#define STI()	        _enable()

class Comm *comm;

Comm::Comm(int b,		// comm port base address
	   int i,		// interrupt number
	   long s,		// BPS rate
	   int l,		// line control setting
	   int fc,		// flow control flag
	   int safety,		// fifo safety (0-3)
	   word si,		// input queue size
	   word so)		// output queue size
{
  word max_size = 64535U / 3;		// max number of input chars
  base = b;				// set up instance base addr
  irq = i;				// ..interrupt request
  oSize = so;				// ..output buffer size
  flow = fc;				// ..flow control flag
  checkFIFO();				// ..fifo flag if available
  if (s)
    setBPS(s, safety);			// ..line speed
  else
    detectBPS(safety);
  if (l)
    setLCR(l);				// ..line format
  if (si < 512)				// q. less than .5k of buffer?
    si = 512;				// a. yes .. set to minimum
  if (si > max_size)			// q. greater than maximum?
    si = max_size;			// a. yes .. set to max size
  if (so < 256)				// q. really small output buf?
    so = 256;				// a. yes .. set to minimum
  iSize = si;				// save input buffer size
  emptyTrans = 1;			// ..set empty transmitter flag
  MSRchanged = 1;			// ..set msr changed flag
  oFlow = iFlow = 0;			// ..clear active flags
  oCount = iCount = 0;		// ..and current queue counts
  iBuf = new char[iSize * 3];		// allocate input buffer
  if (!iBuf) fatalError("memory allocation failed");
  iGet = iBuf;
  iPut = iBuf;				// ..for data, lsr and msr
  iLimit = iBuf + (iSize - 1) * 3;	// ..set limit address
  memset(iBuf, 0, iSize * 3);		// ..and clear area to nulls
  oBuf = oGet = oPut = new char[oSize];	// allocate output buffer
  if (!oBuf) fatalError("memory allocation failed");
  oLimit = oBuf + oSize - 1;		// ..set limit address
  memset(oBuf, 0, oSize);		// ..and clear area to nulls
  setLimits();				// set up flow ctl limits
  installISR();				// ..and interrupt routine
  intMSR = IN(MSR);			// set up initial MSR
}

Comm::~Comm() {
  OUT(MCR, MCR_DTR);		// leave DTR on
  uninstallISR();		// remove interrupt service
  delete iBuf;			// free input
  delete oBuf;			// ..and output queues
}

word Comm::read(char *c) {
  static char dummy;
  return read(c, &dummy, &dummy);
}

word Comm::read(char *c,	// queued character
                char *m,	// modem status register
		char *l)	// line status register
{
  if (iCount == 0)		// q. input queue empty?
    return -1;			// a. yes .. return w/err code

  *c = *iGet++;		// get char from input queue
  *m = *iGet++;		// ..then get modem status
  *l = *iGet++;		// ..then get line status
  iCount--;			// ..finally decrement count

  if (iGet > iLimit)		// q. reached end of buffer?
    iGet = iBuf;		// a. yes .. set to beginning

  if (iFlow && (iCount <= iStart))	// q. port need restarting?
    setRTS(1);				// a. yes .. raise RTS line

  return (((byte) *l == 0xff) ?	// return with character or
      ((*m << 8) + *c) : 0);	// overflow count based on lsr flag
}

void Comm::write(int c) {
  while (!oFree());		// wait until room in queue
  *oPut++ = c;			// put char into output queue
  if (emptyTrans) {		// pump need priming?
    emptyTrans = 0;		// clear empty transmitter flag
    oCount++;			// ..and count character
    if ((!flow || IN(MSR) & MSR_CTS)	// q. output flow satisfied?
	&& (c = deQ()) != -1)		// ..and queue not empty?
      OUT(THR, c);			// a. yes .. send char to port
  } else
    oCount++;				// else .. count characters
  if (oPut > oLimit)			// q. reach end of buffer?
    oPut = oBuf;			// a. yes .. set to beginning
}

void Comm::write(char *s) {
  while (*s) write(*s++);
}

void Comm::write(char *s, int l) {
  while (l--) write(*s++);
}

void Comm::command(char *c, char *r, void (*idle)(int)) {
  char ch, d1, d2;
  iClear();
  oClear();
  while (*c) {
    if ((*c == '^') && (*(c + 1) == 'M')) {
      write('\r');
      if (idle) idle(100); else delay(100);
      c += 2;
    } else if (*c == '~') {
      if (idle) idle(250); else delay(250);
      c++;
    } else if (*c == ' ') {
      c++;
    } else {
      write(*c);
      if (idle) idle(100); else delay(100);
      c++;
    }
  }
  if (r == NULL) return;
  while (!oEmpty());
  while (iEmpty());
  if (idle) idle(500); else delay(500);
  for (int i = 0; i < 127 && !iEmpty(); i++) {
    read(&ch, &d1, &d2);
    r[i] = ch;
  }
  r[i] = 0;
}

word Comm::getIcount() {
  return iCount;			// return nbr in queue
}

word Comm::getOcount() {
  return oCount;			// return nbr in queue
}

word Comm::iFree() {
  return iSize - iCount;		// buffer size less in-use
}

word Comm::oFree() {
  return oSize - oCount;		// buffer size less in-use
}

long Comm::getBPS() {
  return((long) (115200L / (long) divisor));	// return current speed
}

byte Comm::getMSR() {
  return(intMSR);	// return last modem status
}

byte Comm::getLCR() {
  return IN(LCR);
}

int Comm::UART_16550() {
  return fifo;
}

int Comm::carrier() {
  return (intMSR & MSR_CD) != 0;
}

int Comm::getMSRchanged() {
  int rtn = MSRchanged;		// get current status
  MSRchanged = 0;		// ..and clear flag
  return rtn;			// ..and return current status
}

int Comm::getIflow() {
  return iFlow;	// rtn TRUE if input flow has been restricted
}

int Comm::iEmpty() {
  return (iCount == 0) &&		// TRUE if queue empty
	 ((byte) iPut[2] != 0xff);	// ..and not in overflow
}

int Comm::oEmpty() {
  return oCount == 0;			// TRUE if queue empty
}

void Comm::iClear() {
  CLI();			// stop interrupts
  iGet = iPut = iBuf;		// reset buffer pointers
  iCount = 0;			// ..and queue count
  STI();			// ..and re-enable interrupts
}

void Comm::oClear() {
  CLI();			// stop interrupts
  oGet = oPut = oBuf;		// reset buffer pointers
  oCount = 0;			// ..and queue count
  STI();			// ..and re-enable interrupts
}

word Comm::set8N() {
  word oldline = IN(LCR);	// get the current LCR value
  line = ((oldline & LCR_STOP) | LCR_WLEN); // setup 8 databits, no parity
  OUT(LCR, line);		// ..and write new LCR
  return(oldline);		// return old LCR value
}

void Comm::setBPS(long s, int safety) {
  int d = divisor = 115200L / s;
  OUT(IER, 0);			// clear interrupts enable
  if (fifo) {			// q. is UART a 16550 chip?
    safety = 3 - safety;
    OUT(FCR, (safety << 6) | FCR_XMIT | FCR_RCVR | FCR_FIFO);
  }
  IN(LSR);			// read/reset line status reg
  IN(MSR);			// ..and modem status register
  IN(IIR);			// ..and interrupt id register
  IN(RBR);			// ..and receive buffer reg
  CLI();			// stop interrupts
  OUT(LCR, IN(LCR) | LCR_DLAB);		// set divisor latch bit
  OUT(DLM, d >> 8);			// out msb portion of divisor
  OUT(DLL, d & 0xff);			// ..then the lsb portion
  OUT(LCR, IN(LCR) & ~LCR_DLAB);	// clear divisor latch bit
  OUT(MCR, MCR_DO);			// enable DTR, OUT2 and RTS
  STI();				// ..and interrupts
  OUT(IER, IER_RBF | IER_TBE | IER_MSI);	// then enable UART interrupts
}

void Comm::setLCR(int l) {
  line = l & ~LCR_DLAB;		// save new value in instance
  OUT(LCR, line);		// ..and write new LCR
}

void Comm::lowerDTR(word t, void (*idle)(int)) {
  OUT(MCR, MCR_DO & ~MCR_DTR);		// set off DTR control line
  if (idle) idle(t); else delay(t);	// wait a little bit
  OUT(MCR, MCR_DO);			// ..and restore DTR
}

void Comm::setRTS(int f) {
  OUT(MCR, MCR_DO - (f ? 0 : MCR_RTS));	// set mcr register
  iFlow = !f;				// ..and set flow ctl'd flag
}

void Comm::ISR() {
  int c,		// output character
      cnt,		// loop counter
      first_cycle;	// first cycle flag
  char iir,		// interrupt id register
       lsr;		// working line status reg
  while (((iir = IN(IIR)) & IIR_PEND) == 0) {
    switch (iir & IIR_II) {
      case IIR_MSI:			// modem status interrupt
	MSRchanged = 1;			// set msr changed flag
	if ((intMSR = IN(MSR)) & MSR_CTS) {	// CTS set?
	  if (oFlow ||	IN(LSR) & LSR_THRE) {	// fc or THRE empty?
	    oFlow = 0;			// a. yes .. clear flag
	    if ((c = deQ()) == -1)	// q. output queue empty?
	      emptyTrans = 1;		// a. yes .. set flag
	    else
	      OUT(THR, c);		// else .. put out a character
	  }
	}
	continue;                       // ..and check next interrupt

      case IIR_LSI:			// line status interrupt
	intLSR = IN(LSR);		// read line status register
	continue;			// ..and check again

      case IIR_TBE:			// transmitter buffer empty
	cnt = fifo ? 15 : 1;		// set up output fifo size
	first_cycle = 1;		// ..and first cycle flag
	for (; cnt--; ) {
	  if (flow && !(IN(MSR) & MSR_CTS)) {	// receiver not ready?
	    oFlow = first_cycle;	// assert fc if nothing sent
	    break;			// ..and check next interrupt
	  }
	  if ((c = deQ()) == -1) {
	    emptyTrans = 1;
	    break;
	  }
	  OUT(THR, c);			// put out another character
	  first_cycle = 0;		// ..and clear flag
	}
	continue;			// ..and check again

      case IIR_RBF:			// receiver buffer full
	while ((lsr = IN(LSR)) & LSR_DR)	// while data is available
	  enQ(IN(RBR), intMSR, lsr);		// ..get and enQ
	continue;			// check for next interrupt
    }
  }
  outp(I8259, EOI);
}

void Comm::installISR() {
  int mask;			// interrupt mask
  oldComm = _dos_getvect(irq + 8);	// save old comm interrupt rtn
  comm = this;			// ..save instance address
  _dos_setvect(irq + 8, commISR);	// ..establish new routine
  mask = inp(I8259M);		// get current interrupt mask
  mask &= ~(1 << irq);		// determine new mask
  outp(I8259M, mask);		// ..and put in place
}

void Comm::uninstallISR() {
  int mask;			// interrupt mask
  OUT(IER, 0);			// disable UART interrupt
  OUT(FCR, 0);			// ..and fifo, if available
  _dos_setvect(irq + 8, oldComm);	// re-establish old comm rtn
  mask = inp(I8259M);		// get current interrupt mask
  mask |= (1 << irq);		// determine new mask
  outp(I8259M, mask);		// ..and turn off comm int
}

void Comm::detectBPS(int safety) {
  word d;
  OUT(IER, 0);			// clear interrupts enable
  if (fifo) {			// q. is UART a 16550 chip?
    safety = 3 - safety;
    OUT(FCR, (safety << 6) | FCR_XMIT | FCR_RCVR | FCR_FIFO);
  }
  IN(LSR);			// read/reset line status reg
  IN(MSR);			// ..and modem status register
  IN(IIR);			// ..and interrupt id register
  IN(RBR);			// ..and receive buffer reg
  CLI();			// stop interrupts
  OUT(LCR, IN(LCR) | LCR_DLAB);		// set divisor latch bit
  d = (word) IN(DLM) << 8;
  d += (word) IN(DLL);
  OUT(LCR, IN(LCR) & ~LCR_DLAB);	// clear divisor latch bit
  OUT(MCR, MCR_DO);			// enable DTR, OUT2 and RTS
  STI();				// ..and interrupts
  OUT(IER, IER_RBF | IER_TBE | IER_MSI); // then enable UART interrupts
  divisor = d;			// set divisor
}

void Comm::setLimits(void) {
  iStop = (word) (115200L / 10) / divisor;	// get 1 second in characters
  if (iStop > iSize)			// q. limit too high?
    iStop = iSize / 2;			// a. yes .. set at half
  else
    iStop = iSize - iStop;		// else .. set limit point
  iStart = iStop / 2;			// set restart point at 50%
}

void Comm::checkFIFO() {
  fifo = 0;			// assume standard uart
  OUT(FCR, 0xcf);		// try to enable FIFOs
  if ((IN(IIR) & 0xc0) != 0xc0)	// q. FIFO bits found?
    return;			// a. no .. just return
  OUT(FCR, 0);			// turn off FIFOs
  fifo = 1;			// set fifo available flag
}

void Comm::enQ(int c, int m, int l) {
  if (flow && (getIcount() >= iStop)	// q. flow control needed?
      && !iFlow)			// ..and not already on
    setRTS(0);				// a. yes .. clear RTS line
  switch (iFree()) {
    case 0:				// full input queue
      if (iOverflow)				// q. in overflow?
      {					// a. yes .. count lost chars
	if (*(word *) iLast < 65534U)	// q. within lost count range?
	  (*(word *) iLast)++;		// a. yes .. tally another one
      } else {
	iOverflow = 1;			// set overflow flag
	*(word *) iLast = 2;		// init counter to 2
	iLast[2] = 0xff;		// ..and set error flag in lsr
      }
      break;				// ..and return to caller

    case 1:				// almost full
      iLast = iPut;			// save last saved addr
      iOverflow = 0;				// clear overflow flag

    default:
      iCount++;				// count characters in queue
      *iPut++ = c;			// save character in queue
      *iPut++ = m;			// ..and modem status register
      *iPut++ = l;			// ..and line status register
      if (iPut > iLimit)		// q. reach end of buffer?
	iPut = iBuf;			// a. yes .. set to beginning
  }
}

word Comm::deQ(void) {
  char c;			// work character
  if (!oCount)			// q. output queue empty?
    return(-1);			// a. yes .. rtn empty handed
  c = *oGet++;			// get char from output queue
  oCount--;			// show character being removed
  if (oGet > oLimit)		// q. reached end of buffer?
    oGet = oBuf;		// a. yes .. set to beginning
  return c & 0xff;		// ..and rtn with char to send
}

void __interrupt commISR() {
  STI();		// re-enable interrupts
  comm->ISR();		// use object's interrupt rtn
}
