#include <assert.h>
#include <ctype.h>
#include <getopt.h>
#include <ncurses.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <termios.h>
#include "ansi.h"
#include "common.h"
#include "help.h"
#include "output.h"
#include "lgc.h"

void                     ncursesInit();
void                     refreshChannels();
void                     refreshClients(int sendCMD = 1);
void                     request(void *cmdP, int cmdLen);
int                      clientRead();
void                     getMessage(Msg * msg);

static void              init(char *hostname);
static void              deinit();
static int               clientConnect(char *hostname);
static void              ncursesDeInit();
static uid_t             uidFromLogin(char *login);
static void              SIGTERM_Handler(int sig);
static int               command(char *line);
static void              error(char *cmd, int err);
static void              usage(char *usage);
static void              getCommandResponse();

int                      nMsgsWaiting;
Msg                      msg;
struct termios           saveTermios;
char                     defaultMsgName[13];
int                      optDOS = 0;
int                      fd;
int                      nChannels = 0;
int                      curChannel = -1;
ServerInfo               serverInfo;
Channel                 *channel;
int                      clientNo,
                         nClients;
Client                  *client;
WINDOW                  *msgW,
                        *cmdW;

static List              msgList;

int
main(int argc, char **argv)
{
  int                      i,
                           c,
                           x,
                           y,
                           runCmd = 0,
                           len = 0,
                           n_lines = 0;
  char                    *line[MAX_LINES],
                          *cmdline,
                           hostname[64];
  fd_set                   rSet,
                           allSet;

  gethostname(hostname, 64);
  if ((i = getopt(argc, argv, "sd")) != EOF)
    switch (i) {
    case 's':
      strcpy(hostname, argv[optind++]);
      break;
    case 'd':
      optDOS = 1;
      break;
    }
  init(hostname);
  FD_ZERO(&allSet);
  FD_SET(STDIN_FILENO, &allSet);
  FD_SET(fd, &allSet);
  for (i = 0; i < MAX_LINES; i++)
    line[i] = (char *) malloc(MAX_LINELEN);
  cmdline = (char *) malloc(MAX_CMDLEN);
  if (!optDOS)
    wrefresh(cmdW);
  for (;;) {
    rSet = allSet;
    select(fd + 1, &rSet, NULL, NULL, NULL);
    if (FD_ISSET(fd, &rSet))
      while (!clientRead());
    while (nMsgsWaiting) {
      getMessage(&msg);
      processMessage();
    }
    if (FD_ISSET(STDIN_FILENO, &rSet)) {
      if (optDOS)
	c = getchar();
      else
	c = wgetch(cmdW);
      if (isprint(c)) {
	line[n_lines][len++] = c;
	if (!optDOS)
	  waddch(cmdW, c);
      } else if (c == 1) {
	if (!optDOS)
	  _werase(cmdW);
	n_lines = len = 0;
      } else if (c == 4) {
	if (optDOS) {
	  DOSwriteByte(PKT_EXIT);
	  DOSendPacket();
	}
	exit(0);
      } else if (c == 8 || c == 127 || c == 263) {
	if (len) {
	  if (!optDOS) {
	    getyx(cmdW, y, x);
	    wmove(cmdW, y, x - 1);
	    waddch(cmdW, ' ');
	    wmove(cmdW, y, x - 1);
	  }
	  len--;
	} else if (n_lines) {
	  line[n_lines--][0] = '\0';
	  len = strlen(line[n_lines]);
	  if (!optDOS) {
	    _werase(cmdW);
	    waddstr(cmdW, line[n_lines]);
	  }
	} else if (!optDOS)
	  beep();
      } else if (c == 10 || c == 13) {
	runCmd = 1;
      } else if (c == 12) {
	if (!optDOS) {
	  clearok(curscr, TRUE);
	  refresh();
	}
      } else if (isspace(c)) {
	line[n_lines][len++] = ' ';
	if (!optDOS)
	  waddch(cmdW, ' ');
      } else if (!optDOS)
	beep();
      if (len && !(len % COLS)) {
	if (n_lines < MAX_LINES - 1) {
	  line[n_lines][len] = '\0';
	  for (i = len; i >= 0 && line[n_lines][i] != ' '; i--);
	  if (i >= 0) {
	    strcpy(line[n_lines + 1], line[n_lines] + i + 1);
	    while (line[n_lines][i - 1] == ' ' && i)
	      i--;
	    line[n_lines][i] = '\0';
	  } else {
	    line[n_lines + 1][0] = line[n_lines][len - 1];
	    line[n_lines + 1][1] = '\0';
	    line[n_lines][len - 1] = '\0';
	  }
	  n_lines++;
	  len = strlen(line[n_lines]);
	  if (!optDOS) {
	    _werase(cmdW);
	    waddstr(cmdW, line[n_lines]);
	  }
	} else
	  runCmd = 1;
      }
      if (runCmd) {
	runCmd = 0;
	line[n_lines][len] = '\0';
	strcpy(cmdline, line[0]);
	for (i = 1; i <= n_lines; i++) {
	  if (line[i][0] != ' ')
	    strcat(cmdline, " ");
	  strcat(cmdline, line[i]);
	}
	if (command(cmdline)) {
	  if (optDOS) {
	    DOSwriteByte(PKT_EXIT);
	    DOSendPacket();
	  }
	  exit(0);
	}
	if (!optDOS)
	  _werase(cmdW);
	n_lines = len = 0;
      }
      if (!optDOS)
	wrefresh(cmdW);
    }
  }
}

void
init(char *hostname)
{
  int                      cmdLen;
  char                     cmd[MAX_CMDLEN],
                          *s,
                          *s2;
  struct termios           tio;
  struct passwd           *pw;

  if (!isatty(STDIN_FILENO))
    errQuit("lgc: stdin is not a tty.");
  if (getuid() == uidFromLogin("guest"))
    errQuit("lgc: guest denied.");
  fd = clientConnect(hostname);
  listInit(&msgList, NULL, 0, TRUE);
  *(int *) cmd = CMD_CLIENTINFO;
  cmdLen = sizeof(int);
  s = cmd + cmdLen;
  pw = getpwuid(getuid());
  strcpy(s, pw->pw_name);
  cmdLen += strlen(s) + 1;
  s = cmd + cmdLen;
  strncpy(s, pw->pw_gecos, 63);
  s[63] = '\0';
  s2 = s;
  while (*s2 && *s2 != ',')
    s2++;
  *s2 = '\0';
  cmdLen += strlen(s) + 1;
  s = cmd + cmdLen;
  strcpy(s, ttyname(STDIN_FILENO));
  cmdLen += strlen(s) + 1;
  request(cmd, cmdLen);
  if (msg.type != RES_OK)
    errQuit("lgc: client already exists.");
  clientNo = *(int *) msg.data;
  atexit(deinit);
  tcgetattr(STDIN_FILENO, &saveTermios);
  if (optDOS) {
    setvbuf(stdin, NULL, _IONBF, 0);
    setvbuf(stdout, NULL, _IONBF, 0);
    COLS = 80;
    LINES = 25;
    tio = saveTermios;
    cfmakeraw(&tio);
    tcsetattr(STDIN_FILENO, TCSANOW, &tio);
  } else
    ncursesInit();
  signal(SIGTERM, SIGTERM_Handler);
  if (optDOS)
    printf("___LGC_START___");
  msgStr("<C RED><B>LGC v2.12 by Adam McKee</C>\n\n");
  refreshServerInfo();
  msgStr("Type  <B>/help</B>  for online help\n\n");
  writeServerInfo();
  refreshChannels();
  refreshClients();
  paging(1);
  getAutoMsgs();
}

void
deinit()
{
  if (!optDOS)
    ncursesDeInit();
  tcsetattr(STDIN_FILENO, TCSAFLUSH, &saveTermios);
}

int
clientConnect(char *hostname)
{
  int                      fd;
  struct hostent          *hP;
  struct sockaddr_in       addr;

  if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    perrQuit("socket");
  bzero(&addr, sizeof(addr));
  addr.sin_family = AF_INET;
  hP = gethostbyname(hostname);
  if (!hP)
    errQuit("Invalid server: %s", hostname);
  memcpy(&addr.sin_addr, hP->h_addr, hP->h_length);
  addr.sin_port = htons(2001);
  if (connect(fd, (struct sockaddr *) &addr, sizeof(addr)) < 0)
    perrQuit("connect");
  return fd;
}

void
ncursesInit()
{
  initscr();
  start_color();
  cbreak();
  noecho();
  nonl();
  intrflush(stdscr, FALSE);
  keypad(stdscr, TRUE);
  typeahead(-1);
  init_pair(1, COLOR_BLACK, COLOR_WHITE);
  for (int i = 0; i < 8; i++)
    init_pair(i + 2, i, COLOR_BLACK);
  msgW = newwin(LINES - 1, COLS, 0, 0);
  wattr_set(msgW, COLOR_PAIR(COLOR_WHITE + 2));
  _werase(msgW);
  wrefresh(msgW);
  idlok(msgW, TRUE);
  scrollok(msgW, TRUE);
  wsetscrreg(msgW, 0, LINES - 2);
  cmdW = newwin(1, COLS, LINES - 1, 0);
  wattr_set(cmdW, COLOR_PAIR(1));
  _werase(cmdW);
  wrefresh(cmdW);
}

void
ncursesDeInit()
{
  clear();
  refresh();
  endwin();
  ansiAttr(ANSI_NORMALATTR);
}

uid_t
uidFromLogin(char *login)
{
  uid_t                    res = (uid_t) - 1;
  struct passwd           *passwdp;

  if ((passwdp = getpwnam(login)) != NULL)
    res = passwdp->pw_uid;
  return res;
}

void
SIGTERM_Handler(int sig)
{
  sig = sig;
  errQuit("caught a SIGTERM.");
}

int
command(char *line)
{
  int                      quit = 0,
                           cmdLen,
                           li = 0;
  char                    *s,
                           token[MAX_CMDLEN],
                           cmd[MAX_CMDLEN];

  if (getToken(token, line, &li))
    goto l1;
  if (token[0] == '/') {
    if (token[1] == '\0') {
      if (!defaultMsgName[0]) {
	msgStr("<C %d>**</C> No default name defined\n", FG_STAR);
	return quit;
      }
      strcpy(token, defaultMsgName);
      goto l0;
    } else if (!strncasecmp(token, "/automsg", 6)) {
      *(int *) cmd = CMD_AUTOMSG;
      s = cmd + sizeof(int);
      if (getToken(token, line, &li))
	goto usage0;
      if (!strcmp(token, "*"))
	strcpy(s, "*");
      else if (!strcmp(token, "."))
	strcpy(s, channel[curChannel].name);
      else {
	if (token[0] == '#')
	  strcpy(s, token + 1);
	else
	  strcpy(s, token);
      }
      if (getToken(token, line, &li))
	goto usage0;
      s += strlen(s) + 1;
      strcpy(s, token);
      s += strlen(s) + 1;
      if (getToken(token, line, &li))
	goto usage0;
      strncpy(s, token, 12);
      s[12] = '\0';
      s += strlen(s) + 1;
      if (getToken(token, line, &li))
	goto usage0;
      strcpy(s, token);
      while (!getToken(token, line, &li)) {
	strcat(s, " ");
	strcat(s, token);
      }
      s += strlen(s) + 1;
      cmdLen = s - cmd + 1;
      request(cmd, cmdLen);
      if (msg.type != RES_OK)
	error("automsg", msg.type);
      goto l1;
    usage0:
      usage("automsg \\<channel> \\<time> \\<user> \\<message>");
      goto l1;
    } else if (!strncasecmp(token, "/autodel", 6)) {
      *(int *) cmd = CMD_AUTODEL;
      if (getToken(token, line, &li)) {
	usage("autodel \\<prefix>");
	goto l1;
      }
      s = cmd + sizeof(int);
      strcpy(s, token);
      while (!getToken(token, line, &li)) {
	strcat(s, " ");
	strcat(s, token);
      }
      s += strlen(s) + 1;
      cmdLen = s - cmd + 1;
      request(cmd, cmdLen);
      if (msg.type != RES_OK)
	error("autodel", msg.type);
      goto l1;
    } else if (!strncasecmp(token, "/connect", 2)) {
      if (getToken(token, line, &li)) {
	usage("connect \\<user>");
	goto l1;
      }
      *(int *) cmd = CMD_CONNECT;
      strcpy(cmd + sizeof(int), token);
      cmdLen = sizeof(int) +   strlen(token) + 1;
      request(cmd, cmdLen);
      if (msg.type != RES_OK)
	error("connect", msg.type);
    } else if (!strncasecmp(token, "/disconnect", 2)) {
      *(int *) cmd = CMD_DISCONNECT;
      cmdLen = sizeof(int);
      request(cmd, cmdLen);
      if (msg.type != RES_OK)
	error("disconnect", msg.type);
    } else if (!strncasecmp(token, "/exit", 3)) {
      if (curChannel < 0) {
	quit = 1;
	goto l1;
      }
      *(int *) cmd = CMD_LEAVECHANNEL;
      cmdLen = sizeof(int);
      request(cmd, cmdLen);
      curChannel = -1;
      paging(1);
      getAutoMsgs();
      if (optDOS) {
	DOSwriteByte(PKT_CHANNEL);
	DOSwriteString((char *) &curChannel, sizeof(short));
	DOSendPacket();
      }
    } else if (!strncasecmp(token, "/help", 2)) {
      if (!getToken(token, line, &li))
	quit = help(token);
      else
	quit = help(NULL);
    } else if (!strncasecmp(token, "/info", 2)) {
      writeServerInfo();
    } else if (!strncasecmp(token, "/join", 2)) {
      if (getToken(token, line, &li)) {
	usage("join \\<channel>");
	goto l1;
      }
      *(int *) cmd = CMD_JOINCHANNEL;
      if (token[0] == '#')
	strcpy(cmd + sizeof(int), token + 1);
      else
	strcpy(cmd + sizeof(int), token);
      cmdLen = sizeof(int) +   strlen(cmd + sizeof(int)) + 1;
      request(cmd, cmdLen);
      if (msg.type == RES_OK) {
	curChannel = *(int *) msg.data;
	if (optDOS) {
	  DOSwriteByte(PKT_CHANNEL);
	  DOSwriteString((char *) &curChannel, sizeof(short));
	  DOSendPacket();
	}
	refreshClients();
	writeClients(curChannel);
      } else
	error("join", msg.type);
    } else if (!strncasecmp(token, "/list", 2)) {
      if (!getToken(token, line, &li)) {
	if (token[0] == 'c')
	  writeChannels();
	else if (token[0] == '#') {
	  int *chan;

	  *(int *) cmd = CMD_GETCHANNELNO;
	  strcpy(cmd + sizeof(int), token + 1);
	  request(cmd, sizeof(int) + strlen(token));
	  chan = (int*)msg.data;
	  if (*chan < 0) {
	    error("list", RES_ENOCHANNEL);
	    goto l1;
	  }
	  writeClients(*chan);
	} else
	  usage("list [channels|\\<channel>]");
      } else {
	if (curChannel < 0)
	  writeChannels();
	else {
	  refreshClients();
	  writeClients(curChannel);
	}
      }
    } else if (!strncasecmp(token, "/me", 3)) {
      if (getToken(token, line, &li)) {
	usage("me \\<msg>");
	goto l1;
      }
      s = cmd + sizeof(int);
      strcpy(s, token);
      while (!getToken(token, line, &li)) {
	strcat(s, " ");
	strcat(s, token);
      }
      *(int *) cmd = CMD_ACTION;
      strcpy(cmd + sizeof(int), s);
      cmdLen = sizeof(int) + strlen(s) + 1;
      request(cmd, cmdLen);
      if (msg.type != RES_OK)
	error("me", msg.type);
    } else if (!strncasecmp(token, "/msg", 2)) {
      if (getToken(token, line, &li))
	goto usage1;
      strncpy(defaultMsgName, token, 12);
      defaultMsgName[12] = '\0';
    l0:
      *(int *) cmd = CMD_MSGUSER;
      strcpy(cmd + sizeof(int), token);
      s = cmd + sizeof(int) + strlen(token) + 1;
      if (getToken(token, line, &li))
	goto usage1;
      strcpy(s, token);
      while (!getToken(token, line, &li)) {
	strcat(s, " ");
	strcat(s, token);
      }
      cmdLen = sizeof(int) + strlen(cmd + sizeof(int)) + strlen(s) + 2;
      request(cmd, cmdLen);
      if (msg.type != RES_OK)
	error("msg", msg.type);
      goto l1;
    usage1:
      usage("msg \\<user> \\<message>");
      goto l1;
    } else if (!strncasecmp(token, "/nickname", 2)) {
      if (getToken(token, line, &li)) {
	usage("nickname \\<nick>");
	goto l1;
      }
      *(int *) cmd = CMD_NICKNAME;
      token[12] = '\0';
      strcpy(cmd + sizeof(int), token);
      cmdLen = sizeof(int) +   strlen(token) + 1;
      request(cmd, cmdLen);
      if (msg.type == RES_OK)
	strcpy(client[clientNo].name, token);
      else
	error("nickname", msg.type);
    } else if (!strncasecmp(token, "/quit", 2)) {
      quit = 1;
    } else if (!strncasecmp(token, "/start", 2)) {
      *(int *) cmd = CMD_CONNECT;
      strcpy(cmd + sizeof(int), client[clientNo].name);
      cmdLen = sizeof(int) +   strlen(client[clientNo].name) + 1;
      request(cmd, cmdLen);
      if (msg.type != RES_OK)
	error("start", msg.type);
    } else if (optDOS && !strncasecmp(token, "/terminal", 2)) {
      if (getToken(token, line, &li)) {
	usage("terminal \\<cols> [rows]");
	goto l1;
      }
      COLS = atoi(token);
      if (getToken(token, line, &li))
	goto l1;
      LINES = atoi(token);
    } else {
      msgStr("<B>Unknown command</B>: %s\n", token + 1);
    }
  } else {
    if (curChannel >= 0) {
      *(int *) cmd = CMD_MSGALL;
      strcpy(cmd + sizeof(int), line);
      cmdLen = sizeof(int) +   strlen(line) + 1;
      request(cmd, cmdLen);
      msgStr("<B>></B> ");
      msgWrapStr(2, line);
    } else
      msgStr("<C %d>**</C> Global broadcast prohibited\n", FG_STAR);
  }
l1:
  return quit;
}

void
error(char *cmd, int err)
{
  msgStr("<C %d>**</C> %s: ", FG_STAR, cmd);
  switch (err) {
  case RES_EILLEGAL:
    msgStr("illegal operation");
    break;
  case RES_EBADNICK:
    msgStr("nicknames consist of alpha-numeric chars");
    break;
  case RES_ENICKINUSE:
    msgStr("nick is already in use");
    break;
  case RES_ENOUSER:
    msgStr("unknown client");
    break;
  case RES_ENOCHANNEL:
    msgStr("unknown channel");
    break;
  case RES_ENOTLOCAL:
    msgStr("client is not local");
    break;
  case RES_EBUSY:
    msgStr("client is busy");
    break;
  case RES_ENETBUSY:
    msgStr("network is busy");
    break;
  case RES_ENETFULL:
    msgStr("all network nodes are in use");
    break;
  }
  msgStr("\n");
}

void
usage(char *usage)
{
  msgStr("<C %d>**</C> usage: %s\n", FG_STAR, usage);
}

void
refreshServerInfo()
{
  int                      cmd = CMD_SERVERINFO,
                           cmdLen = sizeof(int);

  request(&cmd, cmdLen);
  memcpy(&serverInfo, msg.data, sizeof(ServerInfo));
}

void
refreshChannels()
{
  int                      cmd = CMD_LISTCHANNELS,
                           cmdLen = sizeof(int);

  request(&cmd, cmdLen);
  nChannels = *(int *) msg.data;
  if (channel != NULL)
    free(channel);
  channel = (Channel *) malloc(nChannels * sizeof(Channel));
  memcpy(channel, msg.data + sizeof(int), nChannels * sizeof(Channel));
}

void
refreshClients(int sendCMD)
{
  if (sendCMD) {
    int                      cmd = CMD_LISTCLIENTS;
    request(&cmd, sizeof(int));
  } else
    getCommandResponse();
  nClients = *(int *) msg.data;
  if (client != NULL)
    free(client);
  client = (Client *) malloc(nClients * sizeof(Client));
  memcpy(client, msg.data + sizeof(int), nClients * sizeof(Client));
}

void
request(void *cmdP, int cmdLen)
{
  write(fd, &cmdLen, sizeof(int));
  write(fd, cmdP, cmdLen);
  getCommandResponse();
}

void
getCommandResponse()
{
  for (;;) {
    clientRead();
    if (nMsgsWaiting) {
      getMessage(&msg);
      if (MSG(msg.type))
	processMessage();
      else if (RES(msg.type))
	break;
    }
  }
}

void
getMessage(Msg * msg)
{
  ListNode                *node;

  if (msg->data) {
    delete                   msg->data;

    msg->data = NULL;
  }
  if (!nMsgsWaiting) {
    msg->type = 0;
    msg->len = 0;
    return;
  }
  *msg = *(Msg *) listPeekHead(&msgList, &node);
  listDel(&msgList, node);
  nMsgsWaiting--;
}

int
clientRead()
{
  word                     nRead;
  Msg                     *msg = new Msg;
  struct timeval           tv;
  fd_set                   rset;
  char                     *buf;
 
  tv.tv_sec = 0;
  tv.tv_usec = 0;
  FD_ZERO(&rset);
  FD_SET(fd, &rset);
  if (select(fd + 1, &rset, NULL, NULL, &tv) <= 0) {
    delete msg;
    return 1;
  }

  nRead = 0; buf = (char*)&msg->type;
  while (nRead < sizeof(int))
    nRead += read(fd, buf + nRead, sizeof(int) - nRead);

  nRead = 0; buf = (char*)&msg->len;
  while (nRead < sizeof(int))
    nRead += read(fd, buf + nRead, sizeof(int) - nRead);

  msg->data = new char[msg->len];
  nRead = 0;
  while (nRead < msg->len)
    nRead += read(fd, msg->data + nRead, msg->len - nRead);

  listAddTail(&msgList, msg);
  nMsgsWaiting++;
  return 0;
}
